#ifndef EMS_MATRIX_T1
#define EMS_MATRIX_T1

/*
   NAME: ems_matrixT1.cpp
   SINN:
   //
   INFO:
        2D matrices, organized as a sliced one-dimensional field, have (here so called) minor and major lines:
	if the data atoms of a row build up a sequence in memory the matrix is called RowMajor
	(where as consecutive data atoms of a coloumn will be seperated in memory in this case)

	stepping along the major lines will be faster than along minor lines - reducing the overhead
	in the case of minor lines is the main reason for what will follow;

	to be able to deal with objects acting mostly like real rows or cols is an other reason
	(... but keep in mind: they are handles, providing access to row or col data;
	 they are NOT real rows/cols with there own data)

        g++(gcc-3.2.2-45) limits:
        "sorry, not implemented: use of `enumeral_type' in template type unification"
        instead of:  enum major {}; enum minor {}; template<typename>
          now:  struct major {};  struct minor {};

        ACHTUNG: ISO C++ STL scheint für valarray const keinen Ref sondern value_t zurückzuliefern
        GCC hat dies bislang anders implementiert
*/


#include "ems_matrix_types.h"
//#include "ems_exception.h"

#include <valarray>

/***********************************************************/

namespace EMS {

/////////////////////////////////////////////////////////////


struct DefT1Iterator;
template<class,class> class Matrix_Adapter;

/**/

/* #[ class MatrixT1_base<> : */
template<class T> class MatrixT1_base
{
    template<class,class> friend  class Matrix_Adapter;

public:

    typedef  T              value_t;
    typedef  T *            iterator;
    typedef  T const *      const_iterator;
    typedef  DefT1Iterator  DefIterator_t;


    Size_t                    size( void )  const { return  mData.size(); }

    iterator                 begin( void )        { return  & mData[ 0 ];     }
    const_iterator           begin( void )  const { return  & mData[ 0 ];     }
    iterator                   end( void )        { return  begin() + size(); }
    const_iterator             end( void )  const { return  begin() + size(); }

protected:

                     MatrixT1_base( Size_t  _majorlineS,  Size_t  _minorlineS )
			          :
                                    mData     (_majorlineS * _minorlineS)
                                  , mMajorSize(_minorlineS)
                                  , mMinorSize(_majorlineS)
	                          {}

                     MatrixT1_base( MatrixT1_base const &  m_r )
			          :
                                    mData     (m_r.mData)
                                  , mMajorSize(m_r.mMajorSize)
                                  , mMinorSize(m_r.mMinorSize)
	                          {}

    MatrixT1_base &    operator = ( MatrixT1_base const &  m_r )
                                  {
                                      mData       =  m_r.mData;
                                      mMajorSize  =  m_r.mMajorSize;
                                      mMinorSize  =  m_r.mMinorSize;

                                      return  * this;
                                  }

    void                    resize( Size_t  _majorlineS,  Size_t  _minorlineS )
                                  {
                                      mData.resize( (mMajorSize  =  _minorlineS) * (mMinorSize  =  _majorlineS) );
                                  }

    Size_t               MajorSize( void ) const { return  mMajorSize; }
    Size_t               MinorSize( void ) const { return  mMinorSize; }

    value_t &            unchecked( Size_t  _majorline,  Size_t  _minorline )       { return  mData[ _majorline * mMajorSize + _minorline ]; }
    value_t const &      unchecked( Size_t  _majorline,  Size_t  _minorline ) const { return  mData[ _majorline * mMajorSize + _minorline ]; }

    value_t &              checked( Size_t  _major,  Size_t  _minor )
                                  {
//                                    if( _major >= mMinorSize  ||  _minor >= mMajorSize ) throw BadIdEX("EMS::MatrixT1::at");
                                      return  unchecked( _major, _minor);
				  }
    value_t const &        checked( Size_t  _major,  Size_t  _minor ) const
                                  {
//                                    if( _major >= mMinorSize  ||  _minor >= mMajorSize ) throw BadIdEX("EMS::MatrixT1::at");
                                      return  unchecked( _major, _minor);
                                  }

private:

    mutable  std::valarray<T>  mData;

    Size_t                     mMajorSize;				// in number of (crossing) minor-lines
    Size_t                     mMinorSize;				// in number of (crossing) major-lines
	

    T *                      Start( void )        { return  & mData[ 0 ]; }
    T const *                Start( void )  const { return  & mData[ 0 ]; }
    	                                            
    Size_t                LineSize( major ) const { return  mMajorSize; }
    Size_t                LineSize( minor ) const { return  mMinorSize; }
};
/* #]  : */

/* #[ class MatrixT1RM<> : */
template<class T> class MatrixT1RM : public MatrixT1_base<T>
{
private:

    typedef  MatrixT1_base<T>          base_t;

public:

    typedef  typename base_t::value_t  value_t;
    typedef  major                     row_t;
    typedef  minor                     col_t;


                     MatrixT1RM( Size_t  _rows =0,  Size_t  _cols =0 ) : base_t(_rows, _cols) {}

    void                 resize( Size_t  _rows,  Size_t  _cols )     {         base_t::resize( _rows, _cols );  }

    value_t &        operator()( Size_t  _row,  Size_t  _col )       { return  base_t::unchecked( _row, _col ); }
    value_t const &  operator()( Size_t  _row,  Size_t  _col ) const { return  base_t::unchecked( _row, _col ); }

    value_t &                at( Size_t  _row,  Size_t  _col )       { return  base_t::  checked( _row, _col ); }
    value_t const &          at( Size_t  _row,  Size_t  _col ) const { return  base_t::  checked( _row, _col ); }

    Size_t              RowSize( void ) const { return  base_t::MajorSize(); }
    Size_t              ColSize( void ) const { return  base_t::MinorSize(); }
};
/* #]  : */

/* #[ class MatrixT1CM<> : */
template<class T> class MatrixT1CM : public MatrixT1_base<T>
{
private:

    typedef  MatrixT1_base<T>          base_t;

public:

    typedef  typename base_t::value_t  value_t;
    typedef  major                     col_t;
    typedef  minor                     row_t;


                     MatrixT1CM( Size_t  _rows =0,  Size_t  _cols =0 ) : base_t(_cols, _rows) {}

    void                 resize( Size_t  _rows,  Size_t  _cols )     {         base_t::resize( _cols, _rows );  }

    value_t &        operator()( Size_t  _row,  Size_t  _col )       { return  base_t::unchecked( _col, _row ); }
    value_t const &  operator()( Size_t  _row,  Size_t  _col ) const { return  base_t::unchecked( _col, _row ); }

    value_t &                at( Size_t  _row,  Size_t  _col )       { return  base_t::  checked( _col, _row ); }
    value_t const &          at( Size_t  _row,  Size_t  _col ) const { return  base_t::  checked( _col, _row ); }

    Size_t              RowSize( void ) const { return  base_t::MinorSize(); }
    Size_t              ColSize( void ) const { return  base_t::MajorSize(); }
};
/* #]  : */

/**/

/* #[ struct DefT1Iterator<> : */
struct DefT1Iterator
{
    template<class T> struct traits
    {
	typedef  MatrixT1_base<T>  supported_t;
    };

    /* #[ .class major<T> : */
    template<class T> class major
    {
    public:

	/* #[ class line : */
	class line
	{
	    friend  class DefT1Iterator::major<T>;

	public:

	    typedef  T          value_t;
	    typedef  T *        iterator;
	    typedef  T const *  const_iterator;


	                            line( T * const  t_pc =0,  Size_t  _size =0 )
					:
					  mP   (t_pc)
					, mSize(_size)
	                            {}

	    Size_t                  size( void )        const { return       mSize; }
	    iterator               begin( void )              { return  mP;         }
	    const_iterator         begin( void )        const { return  mP;         }
	    iterator                 end( void )              { return  mP + mSize; }	// begin()  +=  mSize
	    const_iterator           end( void )        const { return  mP + mSize; }

	    value_t &        operator []( Index_t  _i )       { return  mP[ _i ];   }
	    value_t const &  operator []( Index_t  _i ) const { return  mP[ _i ];   }

	private:

	    iterator  mP;
	    Size_t    mSize;	// MajorSize
	};
	/* #]  : */

	typedef  line                           value_t;
	typedef  typename line::iterator        iterator;		// formal nicht korrekt, aber zur Vereinfachung
	typedef  typename line::const_iterator  const_iterator;		//


                        major<T>( void )                                         : mLine(0, 0)        {}
                        major<T>( T * const  t_pc,  Size_t  _size , Index_t =0 ) : mLine(t_pc, _size) {}

	line *    operator ->( void )        const { return  & mLine; }
	line &    operator * ( void )        const { return    mLine; }

	line      operator []( Index_t  _i ) const { return  line(mLine.mP + mLine.mSize * _i, mLine.mSize); }

	major &   operator ++( void )              { mLine.mP  +=  mLine.mSize;      return  * this; }
	major &   operator --( void )              { mLine.mP  -=  mLine.mSize;      return  * this; }
	major &   operator +=( Index_t  _i )       { mLine.mP  +=  mLine.mSize * _i; return  * this; }
	major &   operator -=( Index_t  _i )       { mLine.mP  -=  mLine.mSize * _i; return  * this; }

	major     operator + ( Index_t  _i ) const { return  major<T>(mLine.mP + mLine.mSize * _i, mLine.mSize); }
	major     operator - ( Index_t  _i ) const { return  major<T>(mLine.mP - mLine.mSize * _i, mLine.mSize); }

	bool      operator !=( major<T> const &  l_r ) const { return  mLine.mP != l_r.mLine.mP; }

                  operator T const *      ( void ) const { return  mLine.mP; }
                  operator major<T const> ( void ) const { return  major<T const>(mLine.mP, mLine.mSize); }

    private:

	mutable  line  mLine;
    };
    /* #]  : */

    /* #[ .class minor<T> : */
    template<class T> class minor
    {
    public:

	/* #[ class line : */
	class line
	{
	    friend  class DefT1Iterator::minor<T>;

	public:

	    /* #[ class _iterator<> : */
	    template<class T2> class _iterator
	    {
		friend  class DefT1Iterator::minor<T>;
	
	    private:

		T2 *      mP;
		Index_t   mStep;	// MajorSize

	    public:

		typedef  T2  value_t;


	                       _iterator( T2 * const  t_pc =0,  Index_t  _step =0 )
				   :
			             mP   (t_pc)
				   , mStep(_step)
		               {}

		T2 *         operator ->( void )            const { return     mP;               }
		T2 &         operator * ( void )            const { return  *  mP;               }
		T2 &         operator []( Index_t  _i )     const { return  * (mP + mStep * _i); }

		_iterator &  operator ++( void )                  { mP  +=  mStep;      return  * this; }
		_iterator &  operator --( void )                  { mP  -=  mStep;      return  * this; }
		_iterator &  operator +=( Index_t  _i )           { mP  +=  mStep * _i; return  * this; }
		_iterator &  operator -=( Index_t  _i )           { mP  -=  mStep * _i; return  * this; }

		_iterator    operator + ( Index_t  _i )     const { return  _iterator(mP + mStep * _i, mStep); }
		_iterator    operator - ( Index_t  _i )     const { return  _iterator(mP - mStep * _i, mStep); }

		bool         operator !=( _iterator const &  v_i ) const { return  v_i.mP != mP; }

	                     operator T2 const *         ( void ) const { return  mP; }
		             operator _iterator<T const> ( void ) const { return  _iterator<T const>(mP, mStep); }
	    };
	    /* #]  : */

	    typedef  T                   value_t;
	    typedef  _iterator<T>        iterator;
	    typedef  _iterator<T const>  const_iterator;


	                            line( T * const  t_pc =0,  Index_t  _step =0,  Size_t  _size =0 )
					:
					  mP   (t_pc, _step)
					, mSize(_size)
	                            {}

	    Size_t                  size( void )        const { return                 mSize; }
	    iterator               begin( void )              { return  mP;                   }
	    const_iterator         begin( void )        const { return  mP;                   }
	    iterator                 end( void )              { return  mP + (Index_t) mSize; }	// begin()  +=  mSize
	    const_iterator           end( void )        const { return  mP + (Index_t) mSize; }

	    value_t &        operator []( Index_t  _i )       { return  mP[ _i ]; }
	    value_t const &  operator []( Index_t  _i ) const { return  mP[ _i ]; }

	private:

	    iterator  mP;
	    Size_t    mSize;	// MinorSize
	};
	/* #]  : */

	typedef  line                           value_t;
	typedef  typename line::iterator        iterator;
	typedef  typename line::const_iterator  const_iterator;


                        minor<T>( void )                                             : mLine(0, 0, 0)            {}
                        minor<T>( T * const  t_pc,  Index_t  _step,  Size_t  _size ) : mLine(t_pc, _step, _size) {}

	line *    operator ->( void )        const { return  & mLine; }
	line &    operator * ( void )        const { return    mLine; }
	line      operator []( Index_t  _i ) const { return  line(mLine.mP.mP + _i, mLine.mP.mStep, mLine.mSize); }

	minor &   operator ++( void )              { ++mLine.mP.mP;       return  * this; }	// !!!
	minor &   operator --( void )              { --mLine.mP.mP;       return  * this; }
	minor &   operator +=( Index_t  _i )       { mLine.mP.mP  +=  _i; return  * this; }
	minor &   operator -=( Index_t  _i )       { mLine.mP.mP  -=  _i; return  * this; }

	minor     operator + ( Index_t  _i ) const { return  minor<T>(mLine.mP.mP + _i, mLine.mP.mStep, mLine.mSize); }
	minor     operator - ( Index_t  _i ) const { return  minor<T>(mLine.mP.mP - _i, mLine.mP.mStep, mLine.mSize); }

	bool      operator !=( minor<T> const &  l_i_r ) const { return  l_i_r.mLine.mP.mP != mLine.mP.mP; }

                  operator T const *      ( void ) const { return  mLine.mP.mP; }
	          operator minor<T const> ( void ) const { return  minor<T const>(mLine.mP.mP, mLine.mP.mStep, mLine.mSize); }

    private:

	mutable  line  mLine;
    };
    /* #]  : */
};
/* #]  : */


/////////////////////////////////////////////////////////////

} // namespace EMS

/***********************************************************/

#endif // EMS_MATRIX_T1
